<?php
//
// Report Exporting Utilities
// Copyright (c) 2015-2020 Nagios Enterprises, LLC. All rights reserved.
//


define('EXPORT_PDF', 'pdf');
define('EXPORT_JPG', 'jpg');
define('EXPORT_PORTRAIT', 0);
define('EXPORT_LANDSCAPE', 1);


/**
 * Export a report as a PDF file or JPG image
 *
 * @param   string      $reportname         The name of the report file (i.e. 'availability')
 * @param   constant    $type               Must be EXPORT_PDF (default) or EXPORT_JPG
 * @param   constant    $orientation        Can be either EXPORT_PORTRAIT (default) or EXPORT_LANDSCAPE
 * @param   string      $report_location    Used to override the folder in which the $reportname .php resides
 * @param   string      $filename           The name of the file that runs the report (i.e. index.php or availability.php)
 */
function export_report($reportname, $type = EXPORT_PDF, $orientation = EXPORT_PORTRAIT, $report_location = "reports")
{
    global $cfg;
    $username = $_SESSION['username'];
    $language = $_SESSION['language'];

    // Grab the current URL parts
    $query = array();
    foreach($_GET as $key => $value) {
        $query[$key] = $value;
    }

    // Add in some required components to the query
    $query['token'] = user_generate_auth_token(get_user_id($username));
    $query['locale'] = $language;
    $query['records'] = 100000;
    $query['mode'] = 'getreport';
    $query['hideoptions'] = 1;
    $query['export'] = 1;
    $query['old_browser_compat'] = 1;

    if ($reportname == 'execsummary') {
        $query['records'] = 10;
    }

    $content_type = 'application/pdf';
    if($type == EXPORT_JPG) {
        $content_type = 'application/jpg';
    }

    $url = get_localhost_url() . $report_location . '/' . urlencode($reportname) . '.php?' . http_build_query($query);

    $id = uniqid();
    $tempfile = get_tmp_dir() . "/$reportname-" . $id . ".$type";
    $pid = do_export($url, $reportname, $id, $type, $orientation);
    while(file_exists("/proc/$pid")) {
        usleep(100000);
    }

    if (!file_exists($tempfile)) {
        fail_report($url, $type);
    } else {
        header('Content-type: ' . $content_type);
        header('Content-Disposition: inline; filename="' . time() . '-' . $reportname . '.' . $type . '"');
        readfile($tempfile);
        unlink($tempfile);
    }
}


/**
 * Export multi page report as a PDF file or JPG image.
 *
 * @param   string      $reportname         The name of the report file (i.e. 'availability')
 * @param   int         $record_per_page    Records to be printed per "page". Can expand over multiple pages
 * @param   int         $number_of_pages    Amount of pages to print
 * @param   array       $report_pages       An array of pages to export.
 * @param   constant    $type               Must be EXPORT_PDF (default) or EXPORT_JPG
 * @param   constant    $orientation        Can be either EXPORT_PORTRAIT (default) or EXPORT_LANDSCAPE
 * @param   string      $report_location    Used to override the folder in which the $reportname .php resides
 */
function export_report_multipage($reportname, $record_per_page, $number_of_pages, $type = EXPORT_PDF, $orientation = EXPORT_PORTRAIT, $report_location = "reports") {
    global $cfg;
    $username = $_SESSION['username'];
    $language = $_SESSION['language'];

    // Grab the current URL parts
    $query = array();
    foreach($_GET as $key => $value) {
        $query[$key] = $value;
    }

    // Add in some required components to the query
    $query['locale'] = $language;
    $query['records'] = $record_per_page;
    $query['mode'] = 'getreport';
    $query['hideoptions'] = 1;
    $query['export'] = 1;
    $query['old_browser_compat'] = 1;

    $content_type = 'application/pdf';
    if($type == EXPORT_JPG) {
        $content_type = 'application/jpg';
    }

    # Increment page and add output file to array to connect later
    $url = "";
    $files_to_unite = array();
    $pid_array = array();
    $max_processes = get_option("max_report_processes", 25);
    for($x = 1; $x <= $number_of_pages; $x++) {
        $query['page'] = $x;
        $query['token'] = user_generate_auth_token(get_user_id($username));
        $url = get_localhost_url() . $report_location . '/' . urlencode($reportname) . '.php?' . http_build_query($query);
        $id = uniqid();

        // Collection of pids to wait for. Very bad form of "thread" management
        $pid_array[] = do_export($url, $reportname, $id, $type, $orientation);
        // Collection of filenames for pdfunite
        $files_to_unite[] = get_tmp_dir() . "/$reportname-" . $id . ".$type";

        // Limit amount of background processes
        if(count($pid_array) >= $max_processes) {
            foreach($pid_array as $key => $pid) {
                if (!file_exists("/proc/$pid")) {
                    unset($pid_array[$key]);
                }
            }
        }
    }

    // Wait for each process to finish
    // TODO: Make sure these terminate. Sometimes wkhtmltopdf will hang and consume CPU. Either find root cause or add workaround.
    foreach($pid_array as $pid) {
        while(file_exists("/proc/$pid")) {
            usleep(100000);
        }
    }

    $fail = false;
    foreach($files_to_unite as $file) {
        if (!file_exists($file)) {
            $fail = true;
        }
    }
    if($fail) {
        foreach($files_to_unite as $file) {
            if (file_exists($file)) {
                unlink($file);
            }
        }
        fail_report($url, $type);
    }
     
    $final_report = get_tmp_dir() . "/$reportname-" . uniqid() . ".pdf";
    $reports_united = implode(" ", $files_to_unite);

    `pdfunite $reports_united $final_report`; 

    foreach($files_to_unite as $file) {
        unlink($file);
    }

    if (!file_exists($final_report)) {
        fail_report($url, $type);
    } else {
        header('Content-type: ' . $content_type);
        header('Content-Disposition: inline; filename="' . time() . '-' . $reportname . '.' . $type . '"');
        readfile($final_report);
        unlink($final_report);
    }
}

/**
 * @param   string      $url            URL to pass to wkhtmltox
 * @param   string      $reportname     Name of the report, important for conditional logic
 * @param   int         $id             Unique ID to prevent collision of filenames
 * @param   constant    $type           PDF or JPG
 * @param   constant    $orientation    Landscape or portrait
 * 
 * @return  int      PID of the background process 
 */
function do_export($url, $reportname, $id, $type, $orientation) {
    global $cfg;

    // Do specifics for each type of report
    switch ($type)
    {
        case EXPORT_PDF:
            $bin = 'wkhtmltopdf';
            $opts = '--no-stop-slow-scripts --lowquality --no-outline --footer-spacing 3 --margin-bottom 15mm --footer-font-size 9 --footer-left "' . get_datetime_string(time(), DT_SHORT_DATE_TIME, DF_AUTO, "null") . '"';
            if ($orientation == EXPORT_LANDSCAPE) {
                $opts .= ' -O landscape';
            }
            // These types of reports are collected together with pdfunite so their footers will be wrong.
            if ($reportname !== 'capacityplanning' && $reportname !== 'auditlog') {
                $opts .= ' --footer-right "Page [page] of [toPage]"';
            }
            $ext = '.pdf';
            break;
        case EXPORT_JPG:
            $bin = 'wkhtmltoimage';
            $opts = '--no-stop-slow-scripts';
            $ext = '.jpg';
            break;
        default:
            die(_('ERROR: Could not export report as ') . $type . '. ' . _('This type is not defined.'));
    }

    $delay = get_option('report_export_delay', 400);
    if ($reportname == 'capacityplanning') {
        $delay = get_option('report_export_cp_delay', 1000);
    } 

    $old_browser_compat_jquery1 = get_option('old_browser_compat_jquery1', 1);
    if (!$old_browser_compat_jquery1) {
        $opts .= " --javascript-delay $delay";
    }

    $tempfile = get_tmp_dir() . "/$reportname-" . $id . $ext;
    $log = $cfg['root_dir'].'/var/wkhtmltox.log';

    $cmd = '/usr/bin/' . $bin . " {$opts} ".escapeshellarg($url)." ".escapeshellarg($tempfile) . " &>/dev/null & echo $!";
    
    if (file_exists($log) && is_writable($log)) {
        file_put_contents($log, "[".date(DATE_RFC2822)."] {$cmd}\n", FILE_APPEND);
        // I got rid of outputting to a log file because it gets super messed up with multiple running processes
    }

    $pid = exec($cmd);

    return $pid;
}

function fail_report($url, $type) {
    $color = "#000";
    if (get_theme() == 'xi5dark') { $color = "#EEE"; }
    echo '<div style="margin: 7% auto; color: '.$color.'; max-width: 80%; text-align: center; font-family: verdana, arial; font-size: 1rem; word-wrap: break-word;">';
    echo '<div><strong>' . _('Failed to create ') . '<span style="text-transform: uppercase;">' . $type . '</span></strong></div>';
    echo '<div style="margin: 10px 0 30px 0;">' . _('Verify that your Nagios XI server can connect to the URL') . ':</div>';
    echo '<div style="font-size: 0.7rem;">' . $url . '</div>';
    echo '</div>';
    die();
}
